# Copyright

"""Utilities for chemical inventories.

Includes a CAS number validator and document generation.
"""

import re
import os
import os.path
import time
import types
from sys import stdin, stdout, stderr

from .db.text import DBPrettyPrinter


def valid_CASno(cas_string, debug=False):
    """Validate CAS numbers.

    Check `N..NN-NN-N` format, and the `checksum digit`_ for valid CAS
    number structure.  for

    .. math::
      N_n .. N_4 N_3 - N_2 N_1 - R

    The checksum digit is

    .. math::
      R = remainder([sum_{i=1}^n i N_i  ] / 10 )

    .. _checksum digit:
      http://www.cas.org/expertise/cascontent/registry/checkdig.html

    >>> valid_CASno('107-07-3')
    True
    >>> valid_CASno('107-08-3')
    False
    >>> valid_CASno('107-083')
    False

    Sometimes we don't have a CAS number, or a product will contain
    secret, non-hazardous ingredients.  Therefore we treat the strings
    `na` and `+secret-non-hazardous` as valid CAS numbers.

    >>> valid_CASno('na')
    True
    >>> valid_CASno('+secret-non-hazardous')
    True
    """
    if cas_string in ['na', '+secret-non-hazardous']:
        return True
    # check format,  
    # \A matches the start of the string
    # \Z matches the end of the string
    regexp = re.compile('\A[0-9]{2,}[-][0-9]{2}[-][0-9]\Z')
    if regexp.match(cas_string) == None:
        if debug: print >> stderr, "invalid CAS# format: '%s'" % cas_string
        return False
    # generate check digit
    casdigs = [int(d) for d in ''.join(cas_string.split('-'))]
    sumdigs = casdigs[:-1]
    sumdigs.reverse()
    s = sum([(i+1)*d for i,d in enumerate(sumdigs)])
    check = s % 10
    if check == casdigs[-1]:
        return True
    elif debug:
        print >> stderr, (
            "invalid CAS# check: '%s' (expected %d)" % (cas_string, check))
    return False

class MSDSManager (object):
    """Manage Material Saftey Data Sheets (MSDSs).
    """
    def __init__(self, db, dir="./MSDS/"):
        self.db = db
        self.dir = dir
        self.MIMEs = {
            'application/pdf': ['pdf'],
            'text/html': ['html'],
            'text/plain': ['txt'],
            }
        self.check_dir()

    def check_dir(self):
        "Create the MSDS directory if it's missing."
        if os.path.isdir(self.dir):
            return # all set to go
        elif os.path.exists(self.dir):
            raise Exception, (
                'Error: a non-directory file exists at %s' % self.dir)
        else:
            os.mkdir(self.dir)

    def basepath(self, id):
        assert isinstance(id, int), (
            'id must be an integer, not %s (%s)' % (type(id), str(id)))
        return os.path.join(self.dir, "%d" % id)

    def local_basepath(self, id) : # for symbolic links
        assert isinstance(id, int), (
            'id must be an integer, not %s (%s)' % (type(id), str(id)))
        return "./%d" % id

    def MIME_ext(self, mime):
        if mime in self.MIMEs.keys():
            return self.MIMEs[mime][0]
        for values in self.MIMEs.values():
            if mime in values:
                return mime
        raise ValueError(
            "invalid MIME type '%s'\nshould be one of %s" % (mime, self.MIMEs))

    def path(self, id, mime):
        return "%s.%s" % (self.basepath(id), self.MIME_ext(mime))

    def local_path(self, id, mime):
        return "%s.%s" % (self.local_basepath(id), self.MIME_ext(mime))

    def save(self, id, filetext, mime='application/pdf'):
        "Save the binary byte string FILE to the path for ID"
        print >> file(self.path(id, mime), 'wb'), filetext,

    def link(self, id, target_id):
        # target_id already exists, create a symlink to it for id.
        target_mime = self.get_MSDS_MIME(target_id)
        target_path = self.local_path(target_id, target_mime)
        path = self.path(id, target_mime)
        #os.link(self.path(target_id), self.path(id))   # hard link...
        os.symlink(target_path, path)                   # ... or soft link

    def has_MSDS_MIME(self, id, mime):
        """
        >>> m = MSDSManager(db=None)
        >>> print m.has_MSDS_MIME(102, 'pdf') # test on html
        False
        >>> print m.has_MSDS_MIME(102, 'html') # test on html
        True
        >>> print m.has_MSDS_MIME(6, 'pdf') # test on pdf symlink
        True
        """
        return os.path.exists(self.path(id, mime))

    def get_MSDS_path(self, id):
        """
        >>> m = MSDSManager(db=None)
        >>> print m.get_MSDS_path(102) # test on html
        ./MSDS/102.html
        >>> print m.get_MSDS_path(1) # test on pdf
        ./MSDS/1.pdf
        >>> print m.get_MSDS_path(6) # test on pdf symlink
        ./MSDS/6.pdf
        """
        for mime in self.MIMEs:
            if self.has_MSDS_MIME(id, mime):
                return self.path(id, mime)
        return None

    def get_MSDS_MIME(self, id):
        """
        >>> m = MSDSManager(db=None)
        >>> print m.get_MSDS_MIME(102) # test on html
        text/html
        >>> print m.get_MSDS_MIME(1) # test on pdf
        application/pdf
        >>> print m.get_MSDS_MIME(6) # test on pdf symlink
        application/pdf
        """
        for mime in self.MIMEs:
            if self.has_MSDS_MIME(id, mime):
                return mime
        return None

    def has_MSDS(self, id):
        if self.get_MSDS_path(id) == None:
            return False
        return True

    def get_all(self, simlinks=True):
        ret = []
        for record in self.db.records():
            p = self.get_MSDS_path( int(record['ID']) )
            if p != None:
                if simlinks == False and os.path.islink( p ):
                    continue # ignore the symbolic link
                ret.append({'ID':record['ID'], 'Name':record['Name']})
        return ret


class DocGen (object):
    "Generate the officially required documents"
    def __init__(self, db, doc_root=os.path.join('template', 'doc')):
        self.db = db
        self.doc_root = doc_root

    def _latex_safe(self, string):
        string = string.replace('%', '\%')
        string = string.replace('>', '$>$')
        string = string.replace('<', '$<$')
        return string

    def _set_main_target(self, target):
        print >> file(os.path.join(self.doc_root, 'main.tex'), 'w'), (
            """\documentclass[letterpaper]{article}

\input{%s}
""" % target)

    def _make_pdf(self, target_file):
        os.system('cd %s && make pdf' % self.doc_root)
        path = os.path.join(self.doc_root, target_file)
        os.system('cp %s %s' % (os.path.join(self.doc_root, 'main.pdf'), path))
        return path

    def inventory(self, title=None,
                  namewidth='a', sort_field='db_id',
                  valid_record=lambda r: r['Disposed'] == ''):
        """Create a pdf list of all maching chemicals.  The default is to
        match all currently owned chemicals.  Matching chemicals can be sorted
        by any field (defaults to 'ID')."""
        if title == None:
            title == 'Inventory'
        pp = DBPrettyPrinter(self.db)
        active_ids = []
        for record in self.db.records():
            if valid_record(record) : # get ids for matching chemicals
                active_ids.append(record['db_id'])
        active_ids.sort(cmp=lambda a,b: cmp(self.db.record(a)[sort_field],
                                            self.db.record(b)[sort_field]))
        active_fields = ['ID', 'Name', 'Amount',
                         'H', 'F', 'R', 'O', 'M', 'C', 'T']
        width = {}
        for field in active_fields:
            width[field] = 'a'
        width['Name'] = namewidth
        ## Plain text method
        #string = "Chemical inventory:\t\tGenerated on %s\n\n" \
        #         % time.strftime('%Y-%m-%d')
        #string += pp.multi_record_string(active_ids, active_fields,
        #                                 width=width, FS=' ')
        # return string
        ## Latex method
        string = "\\begin{longtable}{l l l c c c c c c c}\n"
        string += ('%% The header for the remaining page(s) of the table...\n'
                   'ID & Name & Amount & H & F & R & O & M & C & T \\\\\n'
                   '\\hline\n'
                   '\\endhead\n')
        for db_id in active_ids:
            record = self.db.record(db_id)
            string += "  %s & %s & %s & %s & %s & %s & %s & %s & %s & %s \\\\\n" \
                      % (self._latex_safe(record['ID']),
                         self._latex_safe(record['Name']),
                         self._latex_safe(record['Amount']),
                         self._latex_safe(record['H']),
                         self._latex_safe(record['F']),
                         self._latex_safe(record['R']),
                         self._latex_safe(record['O']),
                         self._latex_safe(record['M']),
                         self._latex_safe(record['C']),
                         self._latex_safe(record['T']))
        string += "\\end{longtable}\n"
        print >> file(os.path.join(self.doc_root, 'inventory_title.tex'), 'w'), title
        print >> file(os.path.join(self.doc_root, 'inventory_data.tex'), 'w'), string
        ## alter main.tex to point to the inventory template.
        self._set_main_target('inventory_template')
        ## run latex
        path = self._make_pdf('inventory.pdf')
        return path

    def door_warning(self,
                     valid_record=lambda r: r['Disposed'] == ''):
        """create a warning NFPA diamond and list of the most dangerous
        chemicals for which valid_record(record) is true.  For
        example, to generate a door warning for the front door use
          door_warning(lambda r: r['Disposed'] == '')
        or to generate the warning for the fridge
          door_warning(lambda r: r['Location'] == 'Refrigerator')
        Note that valid_record defaults to the first example.
        """
        pp = DBPrettyPrinter(self.db)
        all_ids = range(self.db.len_records())

        # Search the database to find the nasties
        NFPA_maxs = {'H':0, 'F':0, 'R':0, 'O':[]}
        Mutagens = []
        Carcinogens = []
        Teratogens = []
        Healths = []
        Fires = []
        Reactivities = []
        Others = []
        for record in self.db.records():
            if valid_record(record):
                for field in ['H', 'F', 'R', 'O']:
                    r = record[field]
                    if r != '' and r != '?':
                        if field != 'O' and int(r) > NFPA_maxs[field]:
                            NFPA_maxs[field] = int(r)
                        elif field == 'O' and not r in NFPA_maxs['O']:
                            NFPA_maxs[field].append(r)
                for field,array in zip(['M','C','T'],
                                       [Mutagens,
                                        Carcinogens,
                                        Teratogens]):
                    if record[field] != '' and record[field] != '?':
                        array.append(record['db_id'])
        # now that we've found the max NFPAs,
        # find all the chemicals at those levels
        for record in self.db.records():
            if valid_record(record):
                for field,array in zip(['H', 'F', 'R', 'O'],
                                       [Healths, Fires,
                                        Reactivities, Others]):
                    r = record[field]
                    if r != '' and r != '?':
                        if field != 'O' and int(r) == NFPA_maxs[field]:
                            array.append(record['db_id'])
                        elif field == 'O' and r in NFPA_maxs['O']:
                            array.append(record['db_id'])

        ## generate the output
        # first, update the NFPA grapic code
        if 'OX' in NFPA_maxs['O'] : OX = 'y'
        else :                      OX = 'n'
        if 'W'  in NFPA_maxs['O'] : W  = 'y'
        else :                      W  = 'n'
        os.system('%s %d %d %d %s %s > %s'
                  % (os.path.join(self.doc_root, 'gen_NFPA.sh'),
                     NFPA_maxs['H'], NFPA_maxs['F'], NFPA_maxs['R'], OX, W,
                     os.path.join(self.doc_root, 'mp', 'NFPA.mp')))
        # now generate a list of the nasties ( Amount & ID & Name )
        string = "\\begin{tabular}{r r l}\n"
        for field,name,array in zip(['H', 'F', 'R', 'O'],
                                    ['Health', 'Fire',
                                     'Reactivity', 'Other'],
                                    [Healths, Fires,
                                     Reactivities, Others]):
            if (not hasattr(NFPA_maxs[field], '__len__')) \
                    or len(NFPA_maxs[field]) > 0:
                string += "  \multicolumn{3}{c}{\Tstrut %s : %s} \\\\\n" \
                    % (name, NFPA_maxs[field])
            else : # Print "Other" instead of "Other : []"
                string += "  \multicolumn{3}{c}{\Tstrut %s} \\\\\n" \
                    % (name)
            for db_id in array:
                record = self.db.record(db_id)
                string += "  %s  &  %s &  %s \\\\\n" \
                    % (self._latex_safe(record['Amount']),
                       self._latex_safe(record['ID']),
                       self._latex_safe(record['Name']))
            if len(array) == 0:
                string += "  \multicolumn{3}{c}{ --- } \\\\\n"
        for hazard,array in zip(['Mutagens','Carcinogens','Teratogens'],
                                [Mutagens, Carcinogens, Teratogens]):
            string += "  \multicolumn{3}{c}{\Tstrut %s} \\\\\n" % (hazard)
            for db_id in array:
                record = self.db.record(db_id)
                string += "  %s  &  %s &  %s \\\\\n" \
                    % (self._latex_safe(record['Amount']),
                       self._latex_safe(record['ID']),
                       self._latex_safe(record['Name']))
            if len(array) == 0:
                string += "  \multicolumn{3}{c}{ --- } \\\\\n"                
        string += "\\end{tabular}\n"
        print >> file(os.path.join(self.doc_root, 'door_data.tex'), 'w'), string
        ## alter main.tex to point to the door template.
        self._set_main_target('door_template')
        ## run latex
        path = self._make_pdf('door_warning.pdf')
        return path


def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()
