# Copyright

"""Database-style interface with a text-delimited, single files.

Use this if, for example, your coworkers insist on keeping data
compatible with M$ Excel.
"""

import copy
import os
import os.path
import shutil
import stat
from sys import stdin, stdout, stderr
import time
import types


FILE = 'default.db'
STANDARD_TAB = 8


class IndexStringError (Exception):
    "invalid index string format"
    pass


class FieldError (Exception):
    "database fields are not unique"
    pass


class MissingDatabaseFile (Exception):
    "Specified database file does not exist"
    def __init__(self, path):
        msg = "Missing database file %s" % path
        Exception.__init__(self, msg)
        self.path = path


class TextDB (object):
    """Simple database interface for a spread-sheet style database file.

    field_list()  : return an ordered list of available fields (fields unique)
    long_fields() : return an dict of long field names (keyed by field names)
    record(id)    : return a record dict (keyed by field names)
    records()     : return an ordered list of available records.
    backup()      : save a copy of the current database somehow.
    set_record(i, newvals) : set a record by overwriting any preexisting data
                           : with data from the field-name-keyed dict NEWVALS
    new_record()  : add a blank record (use set_record to change its values)
    """
    def __init__(self, filename=FILE, COM_CHAR='#', FS='\t', RS='\n',
                 current_dir='./current/', backup_dir='./backup/' ):
        self.filename = filename
        self.COM_CHAR = COM_CHAR # comment character (also signals header row)
        self.FS = FS             # field seperator
        self.RS = RS             # record seperator
        self.db_id_field = "db_id" # add a record field for the database index
        # define directories used by the database
        self.cur = current_dir
        self.bak = backup_dir

        if self.filename == None:
            return # for testing, don't touch the file system

        ## Generate the neccessary directory structure if neccessary
        for d in [self.cur, self.bak]:
            self.check_dir(d)

        self._open()

    # directory and file IO operations

    def check_dir(self, dir):
        "Create the database directory if it's missing"
        if os.path.isdir(dir):
            return # all set to go
        elif os.path.exists(dir):
            raise Exception, "Error: a non-directory file exists at %s" % dir
        else:
            os.mkdir(dir)

    def curpath(self):
        "Return the path to the current database file."
        return os.path.join(self.cur, self.filename)

    def _get_mtime(self):
        "Get the timestamp of the last modification to the database."
        s = os.stat(self.curpath())
        return s[stat.ST_MTIME]

    def exists(self):
        "Check if the database exists"
        return os.path.exists(self.curpath())

    def _assert_exists(self):
        """Assert that the database exists on disk.

        Print a reasonable error if it does not.
        """
        if not self.exists():
            raise MissingDatabaseFile(self.curpath())

    def _open(self):
        "Load the database from disk"
        self._assert_exists()
        fulltext = file(self.curpath(), 'r').read()
        self._mtime = self._get_mtime()
        self._parse(fulltext)

    def iscurrent(self):
        "Check if our memory-space db is still syncd with the disk-space db."
        return self._mtime == self._get_mtime()

    def _refresh(self):
        "If neccessary, reload the database from disk."
        if not self.iscurrent():
            self._open()

    def _save(self):
        "Create a new database file from a header and list of records."
        fid = file(self.curpath(), 'w')
        fid.write( self._file_header_string(self._header) )
        if self._long_header:
            fid.write( self._file_header_string(self._long_header) )
        for record in self._records:
            fid.write( self._file_record_string(record) )
        fid.close()

    def backup(self):
        "Back up database file"
        if not self.exists(): return None # nothing to back up.
        # Append a timestamp to the file & copy to self.bak
        tname = '%s.%d' % (self.filename, time.time())
        tpath = os.path.join(self.bak, tname)
        spath = self.curpath()
        shutil.copy(spath, tpath)


    # file-text to memory  operations

    def _get_header(self, head_line, assert_unique=False,
                     assert_no_db_id_field=True):
        """Parse a header line (starts with the comment character COM_CHAR).

        Because doctest doesn't play well with tabs, use colons as field seps.
        >>> db = TextDB(FS=':', filename=None)

        >>> print db._get_header("#Name:Field 1:ID: another field")
        ['Name', 'Field 1', 'ID', ' another field']
        >>> try:
        ...    x = db._get_header("#Name:Field 1:Name: another field", assert_unique=True)
        ... except FieldError, s:
        ...    print s
        fields 0 and 2 both 'Name'
        """
        assert len(head_line) > 0, 'empty header'
        assert head_line[0] == self.COM_CHAR, 'bad header: "%s"' % head_line
        fields = head_line[1:].split(self.FS)
        if assert_unique:
            for i in range(len(fields)):
                for j in range(i+1,len(fields)):
                    if fields[i] == fields[j]:
                        raise FieldError, "fields %d and %d both '%s'" \
                                          % (i,j,fields[i])
        if assert_no_db_id_field:
            for i in range(len(fields)):
                if fields[i] == self.db_id_field:
                    raise FieldError, "fields %d uses db_id field '%s'" \
                                      % (i,fields[i])
        return fields

    def _get_fields(self, line, num_fields=None):
        """Parse a record line.

        Because doctest doesn't play well with tabs, use colons as field seps.
        >>> db = TextDB(FS=':', filename=None)

        >>> print db._get_fields("2-Propanol:4 L:67-63-0:Fisher:6/6/2004",7)
        ['2-Propanol', '4 L', '67-63-0', 'Fisher', '6/6/2004', '', '']
        """
        vals = line.split(self.FS)
        if num_fields != None:
            assert len(vals) <= num_fields, "Too many values in '%s'" % line
            for i in range(len(vals), num_fields):
                vals.append('') # pad with empty strings if neccessary
        return vals

    def _parse(self, text):
        reclines = text.split(self.RS)
        assert len(reclines) > 0, "Empty database file"
        self._header = self._get_header(reclines.pop(0), assert_unique=True)
        self._long_header = None
        self._records = []
        if len(reclines) == 0:
            return # Only a header
        # check for a long-header line
        if len(reclines[0]) > 0 and reclines[0].startswith(self.COM_CHAR):
            self._long_header = self._get_header(reclines.pop(0))
        for recline in reclines:
            if len(recline) == 0 or recline.startswith(self.COM_CHAR):
                continue # ignore blank lines and comments
            self._records.append(self._get_fields(recline, len(self._header)))

    # memory to file-text  operations

    def _file_record_string(self, record):
        """Format record for creating a new database file.

        Because doctest doesn't play well with tabs, use colons as field seps.
        >>> db = TextDB(FS=':', RS=';', filename=None)

        >>> rs = '2-Propanol:4 L:67-63-0:Fisher:6/6/2004:2:3:0'
        >>> db._file_record_string(db._get_fields(rs))
        '2-Propanol:4 L:67-63-0:Fisher:6/6/2004:2:3:0;'
        """
        return "%s%s" % (self.FS.join(record), self.RS)

    def _file_header_string(self, header):
        "Format header for creating a new database file."
        return "%s%s%s" % (self.COM_CHAR, self.FS.join(header), self.RS)


    # nice, stable API for our users

    def field_list(self) :
        "Return an ordered list of available fields (fields unique)"
        return copy.copy(self._header)

    def long_fields(self):
        "Return a dict of long field names (keyed by field names)"
        if self._long_header:
            return dict(zip(self._header, self._long_header))
        else : # default to the standard field names
            return dict(zip(self._header, self._header))

    def record(self, db_id):
        "Return a record dict (keyed by field names)"
        assert isinstance(db_id, int), "id %s not an int!" % str(db_id)
        assert db_id >= 0 and db_id < len(self._records), (
            "record %d does not exist" % db_id)
        d = dict(zip(self._header, self._records[db_id]))
        d['db_id'] = db_id
        return d

    def records(self):
        "Return an ordered list of available records."
        ret = []
        for id in range(len(self._records)):
            ret.append(self.record(id))
        return ret

    def len_records(self):
        "Return len(self.records()), but more efficiently"
        return len(self._records)

    def set_record(self, db_id, newvals, backup=True):
        """Set a record with data from the field-name-keyed dict `newvals`.

        Overwrites any preexisting data.
        """
        if backup:
            self.backup()
        for k,v in newvals.items():
            if k == self.db_id_field:
                assert int(v) == db_id, \
                    "don't set the db_id field! (attempted %d -> %d)" \
                    % (db_id, int(v))
                continue
            # get the index for the specified field
            assert k in self._header, "unrecognized field '%s'" % k
            fi = self._header.index(k)
            # overwrite the record value
            self._records[db_id][fi] = v
        self._save()

    def new_record(self, db_id=None):
        """Create a blank new record and return it.
        """
        record = {}
        for field in self._header:
            record[field] = ''
        record[self.db_id_field] = len(self._records)
        self._records.append(['']*len(self._header))
        return record


class DBPrettyPrinter (object):
    """Define some pretty-print functions for :class:`TextDB` objects.
    """
    def __init__(self, db):
        self.db = db

    def _norm_active_fields(self, active_fields_in=None):
        """Normalize the active field parameter

        >>> from pprint import pprint
        >>> db = TextDB(FS=':', RS=' ; ', filename=None)
        >>> pp = DBPrettyPrinter(db)
        >>> db._parse("#Name:Amount:CAS#:Vendor ; 2-Propanol:4 L:67-63-0:Fisher")

        >>> pprint(pp._norm_active_fields(None))
        {'Amount': True, 'CAS#': True, 'Name': True, 'Vendor': True}
        >>> pprint(pp._norm_active_fields(['Vendor', 'Amount']))
        {'Amount': True, 'CAS#': False, 'Name': False, 'Vendor': True}
        >>> pprint(pp._norm_active_fields('1:3'))
        {'Amount': True, 'CAS#': True, 'Name': False, 'Vendor': False}
        """
        if active_fields_in == None:
            active_fields = {}
            for field in self.db.field_list():
                active_fields[field] = True
            return active_fields
        elif isinstance(active_fields_in, types.StringTypes):
            active_i = self._istr2ilist(active_fields_in)
            active_fields = {}
            fields = self.db.field_list()
            for i in range(len(fields)):
                if i in active_i:
                    active_fields[fields[i]] = True
                else:
                    active_fields[fields[i]] = False
        else:
            if isinstance(active_fields_in, list):
                active_fields = {}
                for field in active_fields_in:
                    active_fields[field] = True
            elif isinstance(active_fields_in, dict):
                active_fields = active_fields_in
            assert isinstance(active_fields, dict), 'by this point, should be a dict'
            for field in self.db.field_list():
                if not field in active_fields:
                    active_fields[field] = False
        return active_fields

    def full_record_string(self, record, active_fields=None):
        """
        Because doctest doesn't play well with tabs, use colons as field seps.
        >>> db = TextDB(FS=':', RS=' ; ', filename=None)
        >>> pp = DBPrettyPrinter(db)
        >>> db._parse("#Name:Amount:CAS#:Vendor ; 2-Propanol:4 L:67-63-0:Fisher")
        >>> print pp.full_record_string( db.record(0) ),
          Name : 2-Propanol
        Amount : 4 L
          CAS# : 67-63-0
        Vendor : Fisher
        """
        fields = self.db.field_list()
        long_fields = self.db.long_fields()
        active_fields = self._norm_active_fields(active_fields)
        # scan through and determine the width of the largest field
        w = 1
        for field in fields:
            if active_fields[field] and len(field) > w:
                w = len(field)
        # generate the pretty-print string
        string = ""
        for field in fields:
            if field in active_fields and active_fields[field]:
                string += "%*.*s : %s\n" \
                          % (w, w, long_fields[field], record[field])
        return string

    def full_record_string_id(self, id, active_fields=None):
        record = self.db.record(id)
        return self.full_record_string(record, active_fields)

    def _istr2ilist(self, index_string):
        """Generate index lists from assorted string formats.

        Parse index strings
        >>> pp = DBPrettyPrinter('dummy')
        >>> pp._istr2ilist('0,2,89,4')
        [0, 2, 89, 4]
        >>> pp._istr2ilist('1:6')
        [1, 2, 3, 4, 5]
        >>> pp._istr2ilist('0,3,6:9,2')
        [0, 3, 6, 7, 8, 2]
        """
        ret = []
        for spl in index_string.split(','):
            s = spl.split(':')
            if len(s) == 1:
                ret.append(int(spl))
            elif len(s) == 2:
                for i in range(int(s[0]),int(s[1])):
                    ret.append(i)
            else:
                raise IndexStringError, "unrecognized index '%s'" % spl
        return ret

    def _norm_width(self, width_in=None, active_fields=None, skinny=True):
        "Normalize the width parameter"
        active_fields = self._norm_active_fields(active_fields)
        if isinstance(width_in, int) or width_in == 'a' : # constant width
            width = {} # set all fields to this width
            for field in active_fields.keys():
                width[field] = width_in
        else:
            if width_in == None:
                width_in = {}
            width = {}
            for field in active_fields.keys():
                # fill in the gaps in the current width
                if field in width_in:
                    width[field] = width_in[field]
                else : # field doesn't exist
                    if skinny : # set to a fixed width
                        # -1 to leave room for FS
                        width[field] = STANDARD_TAB-1
                    else : # set to automatic
                        width[field] = 'a'
        return width

    def _norm_record_ids(self, record_ids=None):
        "Normalize the record_ids parameter"
        if record_ids == None:
            record_ids = range(len(self.db.records()))
        if isinstance(record_ids, types.StringTypes):
            record_ids = self._istr2ilist(record_ids)
        stderr.flush()
        return record_ids

    def _line_record_string(self, record, width=None, active_fields=None,
                               FS=None, RS=None, TRUNC_STRING=None):
        """

        Because doctest doesn't play well with tabs, use colons as field seps.
        >>> db = TextDB(FS=':', RS=' ; ', filename=None)

        >>> pp = DBPrettyPrinter(db)
        >>> db._parse("#Name:Amount:CAS#:Vendor ; 2-Propanol:4 L:67-63-0:Fisher")
        >>> pp._line_record_string_id(0)
        '2-Propa:    4 L:67-63-0: Fisher ; '
        """
        fields = self.db.field_list()
        active_fields = self._norm_active_fields(active_fields)
        width = self._norm_width(width)
        if FS == None:
            FS = self.db.FS
        if RS == None:
            RS = self.db.RS
        for field in fields:
            if field in active_fields and active_fields[field]:
                lastfield = field
        # generate the pretty-print string
        string = ""
        for field in fields:
            if field in active_fields and active_fields[field]:
                w = width[field]
                string += "%*.*s" % (w, w, record[field])
                if field != lastfield:
                    string += "%s" % (FS)
        string += RS
        return string

    def _line_record_string_id(self, id, width=None, active_fields=None,
                               FS=None, RS=None, TRUNC_STRING=None):
        return self._line_record_string(self.db.record(id),
                                        width, active_fields, FS, RS,
                                        TRUNC_STRING)

    def _get_field_width(self, record_ids, field):
        """Return the width of the longest value in FIELD for all the
        records with db_ids in record_ids.
        """
        width = 1
        for i in record_ids:
            w = len(self.db.record(i)[field])
            if w > width:
                width = w
        return width
    def _get_width(self, width_in, active_fields=None, record_ids=None):
        """Return the width of the largest value in FIELD for all the
        records with db_ids in record_ids.
        """
        active_fields = self._norm_active_fields(active_fields)
        width = self._norm_width(width_in, active_fields)
        record_ids = self._norm_record_ids(record_ids)

        for field in active_fields:
            if width[field] == 'a':
                width[field] = self._get_field_width(record_ids, field)
        return width

    def multi_record_string(self, record_ids=None, active_fields=None,
                            width=None, FS=None, RS=None, COM_CHAR=None,
                            TRUNC_STRING=None):
        """

        Because doctest doesn't play well with tabs, use colons as field seps.
        >>> db = TextDB(FS=':', RS=' ; ', filename=None)

        >>> pp = DBPrettyPrinter(db)
        >>> db._parse("#Name:Amount:CAS#:Vendor ; 2-Propanol:4 L:67-63-0:Fisher")
        >>> pp.multi_record_string('0')
        '   Name: Amount:   CAS#: Vendor ; 2-Propa:    4 L:67-63-0: Fisher ; '
        """
        if FS == None:
            FS = self.db.FS
        if RS == None:
            RS = self.db.RS
        if COM_CHAR == None:
            COM_CHAR = self.db.COM_CHAR
        active_fields = self._norm_active_fields(active_fields)
        record_ids = self._norm_record_ids(record_ids)
        width = self._get_width(width, active_fields, record_ids)
        # generate the pretty-print string
        string = ""
        # print a header line:
        fields = self.db.field_list()
        hvals = dict(zip(fields,fields))
        string += "%s" % self._line_record_string(hvals, width,
                                                  active_fields, FS, RS,
                                                  TRUNC_STRING=TRUNC_STRING)
        # print the records
        for id in record_ids:
            string += self._line_record_string_id(id,  width,
                                                  active_fields, FS, RS,
                                                  TRUNC_STRING=TRUNC_STRING)
        return string


def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()
